// cyl_model.h - Generates cylinder models, which are presently used
//   to render cones as well as cylinders.  Modeled after sph_model,
//   but not nearly as weird.

#ifndef __cylmodel_h
#define __cylmodel_h
#pragma once

struct cyl_model {
  float *verts;
  int nverts;

  // scratch data for renderer:
  vertex *proj;
  float *color;

  static cyl_model& get(int sides) {
    static cyl_model* models[100];
    if (sides>=100) sides=99;
    if (!models[sides])
      models[sides] = new cyl_model(sides);
    return *models[sides];
  }

  cyl_model(int sides) {
    nverts = 2*sides;
    verts = new float[3*nverts];
    proj = new vertex[nverts];
    color = new float[4*nverts];

    float *v = verts;
    float a = 0, i = 2*pi/(sides-1);
    for(int s=0;s<sides;s++) {
      float sn = sin(a), cn = cos(a);

      v[0] = 0.0;
      v[1] = sn;
      v[2] = cn;
      v+=3;

      v[0] = 2.0;
      v[1] = sn;
      v[2] = cn;
      v+=3;

      a += i;
    }
  }

  ~cyl_model() {
    delete[] verts;
    delete[] proj;
    delete[] color;
  }
};

#endif