#include "mouseobject.h"

/************* clickObject implementation ***************/

clickObject::clickObject() {
}

void clickObject::init_type() {
  behaviors().name("click");
  behaviors().supportGetattr();
  add_keyword_method("project", &clickObject::py_project);
}

Object clickObject::getattr( const char *_attr ) {
  string attr = _attr;

  if      (attr == "pos")
    return Object(new immutableVector(position));
  else if (attr == "pick")
    return pick;
  else if (attr == "pickpos")
    return Object(new immutableVector(pickpos));
  else if (attr == "camera")
    return Object(new immutableVector(cam));
  else if (attr == "ray")
    return Object(new immutableVector(ray));

  return getattr_methods(_attr);
}

Object clickObject::py_project( const Tuple& args, const Dict& kw ) {
  if (args.length()) throw TypeError("project() requires keyword arguments (name=value)");

  if (!kw.hasKey("normal")) throw TypeError("project: must specify normal=[x y z]");
  Vector normal(kw["normal"]);
  
  double d = 0.0;
  if (kw.hasKey("d")) d = Float(kw["d"]);

  double ndc = normal.dot(cam) - d;
  double ndr = normal.dot(ray);
  if (!ndr) return Nothing();

  double t = -ndc/ndr;
  if (t<0) return Nothing();

  Vector v = cam + ray*t;
  return Object(new immutableVector(v));
}

/************** mouseObject implementation **************/

mouseObject::mouseObject() {
}

void mouseObject::init_type() {
  behaviors().name("mouse");
  behaviors().supportGetattr();
  add_varargs_method("getclick", &mouseObject::py_getclick);
  add_keyword_method("project", &mouseObject::py_project);
}

Object mouseObject::getattr( const char *_attr ) {
  string attr = _attr;

  mutex::lock L(mtx);

  if      (attr == "pos")
    return Object(new immutableVector(position));
  else if (attr == "pick")
    return pick;
  else if (attr == "pickpos")
    return Object(new immutableVector(pickpos));
  else if (attr == "clicked")
    return Int((int)clicks.size());
  else if (attr == "camera")
    return Object(new immutableVector(cam));
  else if (attr == "ray")
    return Object(new immutableVector(ray));

  return getattr_methods(_attr);
}

Object mouseObject::py_getclick(const Tuple& args) {
  if (args.length()) throw TypeError("getclick() expects no arguments");

  while (1) {
    { mutex::lock L(mtx);
      if (clicks.size()) {
        Object c = clicks[0];
        clicks.erase(clicks.begin());
        return c;
      }
    }
    sleep(0.010);
  }
}

Object mouseObject::py_project( const Tuple& args, const Dict& kw ) {
  if (args.length()) throw TypeError("project() requires keyword arguments (name=value)");

  if (!kw.hasKey("normal")) throw TypeError("project: must specify normal=[x y z]");
  Vector normal(kw["normal"]);
  
  double d = 0.0;
  if (kw.hasKey("d")) d = Float(kw["d"]);

  double ndc = normal.dot(cam) - d;
  double ndr = normal.dot(ray);
  if (!ndr) return Nothing();

  double t = -ndc/ndr;
  if (t<0) return Nothing();

  Vector v = cam + ray*t;
  return Object(new immutableVector(v));
}