#ifndef __platwin_h
#define __platwin_h
#pragma once

namespace win {
#include <windows.h>
}

class winCritSec {
  int count;
  win::CRITICAL_SECTION sec;

  public:
  typedef lock<winCritSec> lock;

  winCritSec(int spincount=0, int count=1);
  ~winCritSec();

  inline void sync_lock() { win::EnterCriticalSection(&sec); }
  inline void count_lock() { win::EnterCriticalSection(&sec); count++; }
  inline void sync_unlock() { win::LeaveCriticalSection(&sec); }
  inline int  sync_count() { return count; }

  void clear() { 
    sync_lock();
    count=0;
    sync_unlock();
  }
};

void _threaded_timer(double seconds, bool (*callback)(void*), void *data);

template <class T>
void threaded_timer(double seconds, bool (*callback)(T*), T* data) {
  _threaded_timer(seconds, (bool(*)(void*))callback, (void*)data);
}
  
#endif