// wgl - Windows platform-specific GL and windowing code

#ifndef __WGL_H
#define __WGL_H
#pragma once

#include "pvector.h"

using namespace Py;

#include <map>
#include <string>

using namespace std;

#include "platwin.h"
#include <gl/gl.h>

struct wglContext {
  public:
    void lockMouse();
    void unlockMouse();
    int  getMouseButtons();
    int  getMouseButtonsChanged();
    Vector  getMouseDelta();
    Vector  getMousePos();

    wglContext();
    ~wglContext();

    bool initWindow( const char* title, int x, int y, int width, int height );
    bool changeWindow( const char* title, int x, int y, int width, int height );
    bool isOpen();
    void cleanup();
  
    void makeCurrent();
    void makeNotCurrent();
    void swapBuffers();

    Vector origin();   // of entire window (or, same as initWindow())
    Vector corner();   // of entire window (or, same as initWindow())
    int width();       // of GL area
    int height();      // of GL area

    string lastError() { return error_message; }
    bool   mustPaint;

    static int    sysFont;   // 256 display lists used as a system font
    static void glPrintf(int font, char* format, ...);

  private:
    friend struct wglWindowClass;
    struct wglWindowClass& cls;
    win::HWND  hWnd;
    win::HDC   hdc, lasthdc;
    win::HGLRC hglrc, lastglrc;
    string error_message;

    int buttonState, buttonChanged;
    Vector mousePos, mouseDelta;
    bool mouseLocked;

    win::LRESULT winMessage( win::HWND, win::UINT, win::WPARAM, win::LPARAM );
    void error(string where);
};

typedef wglContext glContext;

#endif