// wgl - Windows platform-specific GL and windowing code

#ifndef __XGL_H
#define __XGL_H
#pragma once

#include "pvector.h"

#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>

using namespace Py;

#include <map>
#include <string>

using namespace std;

#include "platlinux.h"
#include <GL/gl.h>
#include <GL/glu.h>

struct xglContext {
  public:
    void lockMouse();
    void unlockMouse();
    int  getMouseButtons();
    int  getMouseButtonsChanged();
    Vector  getMouseDelta();
    Vector  getMousePos();

    xglContext();
    ~xglContext();

    bool initWindow( const char* title, int x, int y, int width, int height );
    bool changeWindow( const char* title, int x, int y, int width, int height );
    bool isOpen();
    void cleanup();
  
    void makeCurrent();
    void makeNotCurrent();
    void swapBuffers();

    Vector origin();
    Vector corner();
    int width();
    int height();

    string lastError() { return error_message; }
    static int    sysFont;   // 256 display lists used as a system font
    bool   mustPaint;

    static void glPrintf(int font, char* format, ...);

  private:
    GtkWidget *window;
    GtkWidget *area;
    int wwidth, wheight;
    string error_message;

    int buttonState, buttonsChanged;
    Vector mousePos, oldMousePos;
    bool mouseLocked;

    static gint configure_cb(GtkWidget *widget, GdkEventConfigure *event,
                             gpointer data);
    static gint motion_notify_cb(GtkWidget *widget, GdkEventMotion *event,
                                 gpointer data);
    static gint delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
    static gint button_press_cb(GtkWidget *widget, GdkEventButton *event,
                                gpointer data);
    static gint button_release_cb(GtkWidget *widget, GdkEventButton *event,
                                gpointer data);
};

typedef xglContext glContext;

#endif
