from visual import *

print """
Click a ball to select an X slider(red) or an Y slider(green) slider.
Drag with button up, click again to end the drag.
"""

grey = (0.85, 0.85, 0.85)

class slider:
    def __init__(self, pos=vector(0,0,0), axis=vector(0,1.,0), value=None,
                 width=0.15, min=0, max=100., color=(1,0,0), label=""):
        pos = vector(pos)
        axis = vector(axis)
        if value == None:
            value = min
        self.min = min
        self.max = max
        self.value = value
        self.shaft = cylinder(pos=pos, axis=axis, radius=width/4., color=grey)
        self.start = pos
        self.axis = axis
        self.label = label
        self.curcolor = color
        self.width=width
        self.control = sphere(pos=self.start+(self.min+value)/(self.max-self.min)*self.axis,
                        radius=width/2., color = self.curcolor)
        self.control.label = self.label+" = %0.1f" % value
        
    def setcolor(self, newcolor=None):
        if newcolor == None:
            self.control.color = self.curcolor
        else:
            self.control.color = newcolor

    def getslider(self, pos=None):
        pos = vector(pos)
        value = self.min+(self.max-self.min)*dot(pos, self.axis)
        self.setslider(value)
        return value

    def setslider(self, value=None):
        if value > self.max:
            value = self.max
        if value < self.min:
            value = self.min
        self.value = value
        self.control.pos = self.start+(self.min+value)/(self.max-self.min)*self.axis
        self.control.label = self.label+" = %0.1f" % value

class button:
    def __init__(self, pos=vector(0,0,0), value=None, diameter=0.1, color=(1,0,0), label=""):
        pos = vector(pos)
        self.value = value
        self.centre = pos
        self.label = label
        self.curcolor = color
        self.diameter = diameter
        self.button = sphere(pos=self.centre, radius=self.diameter/2., color = self.curcolor)
        self.button.label = self.label

scene.width = 800
scene.height = 400
scene.center = (1,0.5,0)
scene.title = "Testing a Truss"

forever = not(0)    # definition of True 'not zero'

axis = ( (slider(pos=(0,0,0), color=(1,0,0), axis=vector(1.,0,0), max=40., value=5., label="X")),
         (slider(pos=(0,0,0), color=(0,1,0), axis=vector(0,1.,0), max=40., value=5., label="Y")) )

# Testing
buttons =( (button(pos=vector(1.8,.9,0), color = (0,0,1), label="     King Post")),
           (button(pos=vector(1.8,.8,0), color = (0,0,1), label="     Queen Post")),
           (button(pos=vector(1.8,.7,0), color = (0,0,1), label="     Fink")))

while forever:    
    rate(50)
    if scene.mouse.clicked:
        m = scene.mouse.getclick()
        for index in range(2):
            s = axis[index]
            if m.pick is s.control:         # did we click on a sphere?
                pos = m.project(normal=(0,0,1))
                while forever:
                    rate(50)
                    if scene.mouse.clicked:
                        scene.mouse.getclick() # swallow terminating click
                        s.setcolor()
                        break
                    elif scene.mouse.project(normal=(0,0,1)) <> pos:
                        pos = scene.mouse.project(normal=(0,0,1))
                        s.setcolor((1,1,0))
                        s.getslider(pos)
                        continue

        for index in range(3):
            if m.pick is buttons[index].button:         # did we click on a sphere?
                print buttons[index].button.label+" Selected"
                break
        scene.mouse.getclick() # swallow starting click
        print "X = %0.1f, Y = %0.1f" % (axis[0].value,axis[1].value)
