#!/usr/local/bin/perl

package Single_Expression;

&main'handle( 'ANSWER', 'ans tol help', '&Single_Expression\'doAnswer' );
&main'handle( 'RESPONSES', '--', '&Single_Expression\'doResponses' );

#### Handle <ANSWER ...> thing

sub doAnswer {
    local( *attr ) = @_;
    $answer = $main'answer;
    printf "<P><FORM METHOD=GET ACTION=\"%s\">\n", $main'url;
    printf "Answer: <INPUT TYPE=\"text\" NAME=\"ans\" MAXLENGTH=\"40\"";
    printf " VALUE=\"$answer\"" if $answer;
    printf ">\n";
    printf "<INPUT TYPE=\"submit\" VALUE=\"Submit Answer\">\n";
    printf "</FORM>\n";
    if( $attr{'help'} ) {
        printf "<P><A HREF=\"%s\">Help</A> is available.\n", $attr{'help'};
    }
    $main'correctAnswer = $attr{'ans'};
    $tolerance = defined($attr{'tol'}) ? $attr{'tol'} : 0.0001;
##  $epsilon = defined($attr{'eps'}) ? $attr{'eps'} : -1;
    $main'answeredCorrectly = &isCorrect( $answer );
}


#### Handle the RESPONSES section.  If an answer was not supplied,
#### then skip everything.  Otherwise, analyze the WHEN clauses to
#### determine if any output is to be produced.  WHEN CORRECT and
#### WHEN INCORRECT lines are defaults to be output only if no other
#### lines have been.

sub doResponses {
	$answer = $main'answer;
    print "<HR>\nYour answer was: $answer<p>\n" if $answer;
    local($printing,$printed,%attr);
    while( <main'IN> ) {
        last if m|^\s*<\s*/RESPONSES\s*>|i;
        next if !$answer;
        if( &main'isElement('WHEN',*attr, 'ans expr correct incorrect' ) ) {
            $printing = 0;
            if( $attr{'ans'} ) {
                $printing = &isEqual($answer,$attr{'ans'});
            } elsif( $attr{'correct'} ) {
                $printing = $main'answeredCorrectly && !$printed;
            } elsif( $attr{'incorrect'} ) {
                $printing = !$main'answeredCorrectly && !$printed;
            } elsif( $attr{'expr'} ) {
                $printing = &evalExpr( $attr{'expr'} );
            }
            print "<p>" if $printing;
        }
        ($printed = 1), print if $printing;
    }
    &main'defaultRespons() if $answer && !$printed;
	$main'responsePrinted = 1;
}

sub isCorrect {
    local( $ans ) = @_;
    $ans && &isEqual( $ans, $main'correctAnswer );
}

sub isEqual {
    local( $ans1, $ans2 ) = @_;
    $rc = &compare( $ans1, $ans2, $tolerance );
#    print "$ans1 == $ans2: $rc\n";
    return $rc;
}

sub evalExpr {
    local( $e ) = @_;
    $e =~ s/([a-z][a-z0-9]*)/$main'query{$1}/eg;
    local( $rc );
    $rc = eval $e;
#   print $e => $rc\n";
    return $@ ? 0 : $rc;
}

%funs = ( 'atan2',2,  'cos',1,  'exp',1,  'log',1,  'sin',1,  'sqrt',1 );

# compare the two argument expressions for equality.  Return 1 if equal
# within a tolerance given by the third argument.  Return 0 if not equal
# or if something is wrong with expressions (bad syntax, different variables,
# etc.).  In the latter case, set $@ to a string indicating the cause of
# the failure to compare.  Expressions are compared by substituting random
# numbers in the range 0 to 1 for the variables and evaluating numerically.
# Expressions are normal expressions except that we allow things like
# "3i", which is translated to "3*i".  In other words, a number followed
# by an identifier is taken to mean multiplication.

sub compare {
    local( $a, $b, $tol ) = @_;
    $tol = 0.0001 if !$tol;
    $@ = '';
                # translate the expressions and get the variable set for each.
    local( $at, $bt, %av, %bv, $v );
    return &fail('bad syntax',$a) if !($at = &translate( $a, *av ));
    return &fail('bad syntax',$b) if !($bt = &translate( $b, *bv ));
                # ensure that each has the same set of variables.
    foreach $v (keys %av) {
        return &fail('different variables',$a) if !$bv{$v};
    }
    foreach $v (keys %bv) {
        return &fail('different variables',$a) if !$av{$v};
    }
                # substitute random numbers for the identifiers
    foreach $v (keys %av) {
        $av{$v} = rand;
    }
    $at =~ s/\$([a-zA-Z]\w*)/$av{$1}/g;
    $bt =~ s/\$([a-zA-Z]\w*)/$av{$1}/g;
                # evaluate the expressions
    local( $ar, $br );
    $ar = eval $at;
    return &fail("bad syntax",$a) if $@;
    $br = eval $bt;
    return &fail("bad syntax",$b) if $@;
                # compare the numerical result
    return 1 if &isClose( $ar, $br, $tol );
    return &fail();
}
    
# translate the argument expression and build an associative array
# containing the variables.  Return the translated expression or ""
# if the expression is invalid.  In the latter case, set $@ to say
# why.  translate consists of changing occurences of "number
# identifier" to "number * identifier", putting a '$' in front of
# variables, and checking to ensure that only allowable functions
# have been called (i.e., the perl numeric functions).

sub translate {
    local( $e, *symtab ) = @_;
    %symtab = ();
    local( $r, $num, $var );
    while( $e ) {
                    # strip leading non-ident, non-numbers
        $num = 0, $r .= $1 if $e =~ s/^(\s*[^A-Za-z0-9]+\s*)//;
        last if !$e;
                    # is the next thing a number?
        if( $e =~ s/^(\s*\d+(\.?\d*([eE][-+]?\d+)?)?)// ) {
            $r .= '*' if $num;
            $num = 1;
            $r .= $1;
                    # is the next thing an identifier?
        } elsif( $e =~ s/^(\s*([a-zA-Z]\w*))// ) {
            $var = $2;
            $r .= '*' if $num;
            $num = 0;
            if( $e =~ /^\(/ ) { # function?
                ($@ = 'invalid function'), return "" if !$funs{$var};
                $r .= $var;
            } else {            # no, so therefore variable
                $r .= "\$$var";
                $symtab{$var} = 1;
            }
        } else {
            $@ = "invalid token $e";
            return "";
        }
    }
    return $r;
}

# compare two floating point values for almost equality

sub abs { local($n) = @_; return $n >= 0.0 ? $n : -$n; }
sub max { local($a,$b) = @_; return $a >= $b ? $a : $b; }

sub isClose {
    local( $a, $b, $tol ) = @_;
    local( $maxabs );
    $maxabs = &max( &abs($a), &abs($b) );
    return 1 if $maxabs == 0;
    &abs( $a - $b ) <= $tol * $maxabs;
}

# set $@ to a reson string and return 0

sub fail {
    local( $str, $expr ) = @_;
    $reason = $str;
    $reason .= "($@)" if $@;
    $reason .= ": $expr" if $expr;
    $@ = $reason;
    return 0;
}

1;
